/**
 *
 * Programmbeschreibung:
 * ---------------------
 * Es handelt sich um einen vereinfachten WEB-Server, der nach dem dem Hypertext 
 * Transfer Protocol Version 1.0 arbeitet und nur den Befehl GET versteht. Wird 
 * ein anderer Befehl gesendet, so bleibt dieser unbeantwortet. 
 * Dieser Server wertet nur die gesendete URL aus. Header-Zeilen, wie zum Beispiel 
 * Accept, If-Modified-Since, werden nicht bercksichtigt. 
 *
 * @author  Peter Ren Dietmller
 * @version 1.0, 01/99
 * 
 */

import java.io.*;
import java.net.*;
import java.util.*;
import java.text.*;


/**
 * Thread WWWVerbindung
 * --------------------
 * Fr jede eingegangene Verbindung wird ein Exemplar der Klasse WWWVerbindung 
 * angelegt, die sich um die Anfrage kmmert.
 */
class WWWVerbindung extends Thread {
 
  /* -- Datumsformat -- */
  static DateFormat df = DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.MEDIUM, Locale.UK);

  Socket s;        /* Socket, auf dem die Anfrage kam   */
  String rootDir;  /* Wurzelverzeichnis des WWW-Servers */

  public WWWVerbindung(Socket s, String rootDir) {
    this.s = s;
    this.rootDir = rootDir;
  }
  
  private void PrintLog(String ip, int port, String line) {
    System.out.println(ip + ":" + port + ", " + line);
  }
  
  /* sendet den Header fr ein HTML-Objekt */
  private void SendHeader(PrintWriter sout, String rsp, String ctt, long ctl, Date lm) {
    sout.println("HTTP/1.0 " + rsp);
    sout.println("Server: PDI/0.1Beta");
    if (!ctt.equals(""))
      sout.println("Content-Type: " + ctt);
    sout.println("Content-Length: " + ctl);
    sout.println("Date: " + df.format(new Date()));
    if (lm != null)
      sout.println("Last-Modified: " + df.format(lm));
    sout.println();
    sout.flush();
  }
  
  /* sendet den Inhalt einer Datei */
  private void SendFile(File f, OutputStream os) {

    int             len;
    byte[]          data;
    FileInputStream fin;
    
    /* -- Initialisieren -- */
    fin  = null;
    data = new byte[1024];

    try {
      /* -- InputStream anlegen -- */
      fin = new FileInputStream(f);
      /* -- Daten kopieren -- */
        while ((len = fin.read(data)) > 0) os.write(data, 0, len);
    } catch(FileNotFoundException e) {
    } catch(IOException e) {
    } finally {
      try {
        if (fin != null) fin.close();
      } catch(IOException ioe) {
      }
    }
      
  }
  
  /* sendet eine Fehlermeldung an den Client */
  private void SendError(PrintWriter sout, int error) {
  
    String errMsg;
    
    /* -- Fehlermeldung festlegen -- */
    switch(error) {
      case 200: errMsg = "OK";                    break;
      case 404: errMsg = "Objekt nicht gefunden"; break;
      default:  errMsg = "unbekannter Fehler";    break;
    }

    /* -- Header -- */
    SendHeader(sout, error + " " + errMsg, "", 0, null);
    
    /* -- Fehlermeldung -- */
    sout.println("<HTML>");
    sout.println("<HEAD>");
    sout.println("<TITLE>Fehler " + error + ": " + errMsg + "</TITLE>");
    sout.println("</HEAD>");
    sout.println("<BODY>");
    sout.println("<H1>Fehler " + error + ": " + errMsg + "</H1>");
    sout.println("</BODY>");
    sout.println("</HTML>");
    sout.flush();

  }
  
  public void run() {
  
    /* -- HTTP-Befehle -- */
    final int CMD_GET  = 1;
    final int CMD_POST = 2;
  
    String ctt;  /* Content-Type         */
    int    cmd;  /* gesendeter Befehl    */
    String ext;  /* Extension d. lok. N. */
    String fn;   /* lokaler Dateiname    */
    File   f;    /* lokale Datei         */
    String ip;   /* IP Adresse d. Client */
    String line; /* gesendete Zeile      */
    Date   lm;   /* letzte nderung (last modified) der lokalen Datei */
    int    port; /* Port d. Client       */
    String url;  /* gewnschte URL       */
    
    /* -- Variablen initialisieren -- */
    cmd  = 0;
    ip   = "";
    port = 0;
    url  = "";
  
    try {
  
      /* -- Variablen setzen -- */
      ip   = this.s.getInetAddress().getHostAddress();
      port = this.s.getPort();

      /****************************/
      /* Gesendete Daten einlesen */
      /****************************/
      BufferedReader sin = new BufferedReader(new InputStreamReader(this.s.getInputStream()));
      line = sin.readLine();
      while (line.length() > 0) {
          
        /* -- Trace ausgeben -- */
        //PrintLog(ip, port, line);

        /* -- Verarbeite GET-Befehl -- */
        if (line.startsWith("GET")) {
          /* -- Befehl merken -- */
          cmd = CMD_GET;
          /* -- URL ermitteln -- */
          url = "";
          int endOfUrl = line.indexOf(' ', 4);
          if (endOfUrl > 4) url = line.substring(4, endOfUrl);
        }

        line = sin.readLine();

      }

      /*********************************/
      /* Gesendeten Befehl verarbeiten */
      /*********************************/
      if (cmd == CMD_GET) {
      
        /* -- Variable url enthlt den angeforderten URL -- */
        PrintLog(ip, port, "GET " + url);
      
        /* -- lokalen Dateinamen ermitteln -- */
        f = new File(this.rootDir + url);
        if (f.isDirectory()) f = new File(f + "default.htm");
        
        /* -- Extension ermitteln -- */
        ext = "";
        fn  = f.getAbsolutePath();
        int startOfExt = fn.lastIndexOf('.');
        if ((startOfExt > fn.lastIndexOf('/')) && (startOfExt > 0))
          ext = fn.substring(startOfExt + 1);
        /* Anmerkung: Wenn der Dateiname zum Beispiel ab.cd/xy lautet, dann bekommt
         * startOfExt den Wert 2 und als Extension wrde cd/xy herauskommen, was
         * natrlich falsch ist. Die zustzliche Bedingung 
         * startOfExt > lastIndexOf('/') verhindert, da eine Extension eines 
         * Verzeichnisnamens herangezogen wird. */
        
        /* -- Content-Type ermitteln -- */
        ctt = "";
        if (ext.equalsIgnoreCase("htm") || ext.equalsIgnoreCase("html")) 
          ctt = "text/html";
        else if (ext.equalsIgnoreCase("gif"))                                 
          ctt = "image/gif";
        else if (ext.equalsIgnoreCase("jpg") || ext.equalsIgnoreCase("jpeg"))
          ctt = "image/jpeg";
        else if (ext.equalsIgnoreCase("txt") || ext.equalsIgnoreCase("text"))
          ctt = "text/plain";
        else if (ext.equalsIgnoreCase("exe"))
          ctt = "application/x-msdownload";

        /* -- HTML-Seite zurcksenden -- */
        PrintWriter sout = new PrintWriter(this.s.getOutputStream());
        if (f.exists() && f.canRead()) {
          lm = new Date(f.lastModified());
          SendHeader(sout, "200 OK", ctt, f.length(), lm);
          SendFile(f, this.s.getOutputStream());
          PrintLog(ip, port, f.getAbsolutePath() + " gesendet.");
        } else { /* HTML-Objekt existiert nicht */
          SendError(sout, 404);
        }
        
      }
      
    } catch(IOException e) {
      System.out.println(ip + ":" + port + " IOException");
      
    } finally {
      /* -- Verbindung schlieen -- */
      try {
        this.s.close();
      } catch(IOException ioe) {
      }
    }
    
  }

}


public class WWWServer {

  /* -- Konstanten -- */
  final static int PORT = 80;
  
  /* -- Hauptprogramm -- */
  public static void main(String[] args) {
  
    ServerSocket ss = null;     /* Server Socket                 */
    Socket       cs = null;     /* Client Socket                 */
    String       rootDir = "";  /* Wurzelverzeichnis des Servers */
    
    /* -- Startmeldung ausgeben -- */
    System.out.println();
    System.out.println("WWWServer - Einfacher WWW-Server");
    System.out.println("(C)opyright by Peter Ren Dietmller, 1998-99");
    System.out.println();
    
    /* -- Parameter verarbeiten -- */
    if (args.length != 1) {
      System.out.println("Sie haben keinen oder zu viele Parameter angegeben.");
      System.out.println();
      System.out.println("Aufruf: java WWWServer <RootDirectory>");
      System.out.println();
      return;
    }
    rootDir = args[0];
    if (!(new File(rootDir)).exists()) {
      System.out.println("Das Verzeichnis " + rootDir + " existiert nicht.");
      System.out.println();
      return;
    }

    /* -- Server starten -- */
    System.out.println("Starte WWW-Server auf Port " + PORT + " ... (mit Strg+C beenden)");
    
    try {
      /* -- Warte auf eingehende Verbindungen -- */
      ss = new ServerSocket(PORT);
      while (true) {  /* Endlosschleife */
        cs = ss.accept();
        (new WWWVerbindung(cs, rootDir)).start();
      }
      
    } catch(IOException e) {
      System.out.println("Port " + PORT + " ist bereits belegt.");
      
    } finally {
      if (ss != null) {
        try { 
          ss.close(); 
        } catch(IOException ioe) {
        }
      }
      System.out.println("Beende WWW-Server ...");
      
    }  
    
  }

}
